import struct
try:
	from cStringIO import StringIO as BytesIO
except ImportError:
	from io import BytesIO
try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib
	
		
class FakeParser(object):
	def __init__(self, target):
		self.target = target

	def feed(self, data):
		self.target.feed(data)

	def close(self):
		pass


class cachable_xmlrpc_buffer:
	def __init__ (self, cache = 0):
		self.parser, self.buf = xmlrpclib.getparser()
		self.fp = BytesIO ()
		self.cache = cache
		self.cdata = None
	
	def __del__ (self):
		self.fp.close ()
		
	def feed (self, data):
		self.fp.write (data)
		self.parser.feed (data)
				
	def close (self):
		if self.cdata:
			return self.cdata
		res = self.buf.close ()
		if self.cache:
			self.cdata = res		
		return res
	
	def raw (self):
		self.fp.seek (0)
		return self.fp
		
	def no_cache (self):
		self.cache = 0
		self.cdata = None
		
	
class list_buffer:
	def __init__(self, cache = 0):
		self.cache = cache
		self.data = []
		self.cdata = b''
		self.fp = None
	
	def __del__ (self):
		if self.fp:
			self.fp.close ()
			self.fp = None
			
	def __len__ (self):
		return len (self.data)
		
	def feed(self, data):
		if self.cache:
			self.cdata += data
		self.data.append (data)
	
	def raw (self):
		if self.cdata:
			self.fp = BytesIO (self.cdata)			
		else:
			self.fp = BytesIO (b"".join (self.data))
		self.fp.seek (0)
		return self.fp
		
	def read (self):	
		# consume data, used by proxy response
		data = b''.join(self.data)
		self.data = []
		return data
		
	def build_data (self):
		if self.cdata:
			return self.cdata
		return b''.join(self.data)
	
	def close (self):
		if self.cdata:
			return self.cdata
		return self.build_data ()		
	
	def no_cache (self):
		self.cache = 0
		self.cdata = b''


class bytes_buffer:
	def __init__(self, cache = 0):
		self.cache = cache
		self.fp = BytesIO ()
		self.current_buffer_size = 0
		self.cdata = None
	
	def __del__ (self):
		self.fp.close ()		
			
	def __len__ (self):
		return self.current_buffer_size
	
	def raw (self):
		self.fp.seek (0)
		return self.fp
				
	def feed (self, data):
		self.current_buffer_size += len (data)
		self.fp.write (data)
			
	def build_data (self):
		return self.fp.getvalue ()
	
	def close (self):
		if self.cdata:
			return self.cdata
		res = self.build_data ()
		if self.cache:
			self.cdata = res		
		return res
	
	def no_cache (self):
		self.cache = 0
		self.cdata = []
		
def getfakeparser (target_class, cache = False):
	target = target_class (cache)
	return FakeParser (target), target




import sys, os
from . import pathtool
import importlib
import importlib.machinery

try: _reloader = importlib.reload
except AttributeError: _reloader = reload

def importer (directory, libpath, module = None):
	fn = not libpath.endswith (".py") and libpath + ".py" or libpath
	modpath = os.path.join (directory, fn)
	hname = fn.split (".")[0]
	p = directory.replace ("\\", "/")
	if p.find (":") !=- 1:
		p = "/" + p.replace (":", "")
	while 1:
		if "skitai.mounted." + hname in sys.modules:				
			p, l = os.path.split (p)
			if not l:
				raise SystemError ('module %s already imported, use reload')
			hname = l + "." + hname
		else:
			hname = "skitai.mounted." + hname
			break
	
	loader = importlib.machinery.SourceFileLoader(hname, modpath)
	mod = loader.load_module ()
	
	return mod, mod.__file__
	
def reimporter (module, directory, libpath):
	del sys.modules [module.__name__]		
	try: 
		return importer (directory, libpath)
	except:	
		sys.modules [module.__name__] = module
		raise
	

#----------------------------------------------
# will be deprecated
#----------------------------------------------
	
def importer_old (directory, libpath):	
	sys.path.insert(0, directory)	
	__import__ (libpath, globals ())
	libpath, abspath = pathtool.modpath (libpath)
	module = sys.modules [libpath]	
	if abspath [-4:] in (".pyc", ".pyo"):
		abspath = abspath [:-1]		
	sys.path.pop (0)		
	return module, abspath

def reloader (module):
	directory = os.path.split (module.__file__) [0]
	sys.path.insert(0, directory)
	_reloader (module)
	sys.path.pop (0)		



import sqlite3
import json
import zlib

class open:
	def __init__ (self, path):		
		self.conn = sqlite3.connect (path, check_same_thread = False)
		self.create_cursor ()
	
	def create_cursor (self):	
		self.cursor = self.c = self.conn.cursor ()
	
	def to_safe_tuple (self, karg):
		revised = []
		for k, v in karg.items ():
			if isinstance (v, sqlite3.Binary):
				pass
			elif k in (None, ''):
				v = 'null'
			elif isinstance (v, str):
				v = "'{}'".format (v.replace ("'", "''"))
			else:
				v = str (v)
			revised.append ((k, v))
		return revised	
	
	def to_tuple (self, karg):
		revised = []
		for k, v in karg.items ():			
			revised.append ((k, v))
		return revised	
		
	def patch (self, tbl, karg, cond = None):
		tp = self.to_tuple (karg)
		sql = "update {} {}".format (
			tbl,
			",".join (["%s=?".format (k, v) for k, v in tp]),
			cond and "where " + cond or ""
		)
		self.c.execute (sql, tuple ([v for k, v in tp]))
		
	def post (self, tbl, karg):
		tp = self.to_tuple (karg)
		sql = "insert into {} ({}) values ({})".format (
			tbl,
			",".join ([k for k, v in tp]),
			",".join (["?"] * len (tp))
		)		
		self.c.execute (sql, tuple ([v for k, v in tp]))
	
	def __enter__ (self):
		return self
	
	def __exit__ (self, type, value, tb):
		self.c.close ()
		self.conn.close ()
		
	def __getattr__ (self, name):
		return getattr (self.c, name)
	
	def commit (self):	
		return self.conn.commit ()
	
	def rollback (self):	
		return self.conn.rollback ()	

	def serialize (self, obj):
		return zlib.compress (json.dumps (obj).encode ("utf8"))	
	
	def deserialize (self, data):
		return json.loads (zlib.decompress (data).decode ('utf8'))
	
	def blob (self, obj):
		return sqlite3.Binary (obj)
		



from . import response as http_response
from ..http2 import request_handler as http2_request_handler
from . import base_request_handler, http_auth, respcodes
from ...client import asynconnect
import base64
from hashlib import md5
import os
from hyperframe.frame import SettingsFrame
from ..http2 import H2_PROTOCOLS
from ...client import socketpool
from ...lib.cbutil import tuple_cb
import asyncore

class ProtocolSwitchError (Exception):
	pass


class RequestHandler (base_request_handler.RequestHandler):
	FORCE_HTTP_11 = False	
	
	def __init__ (self, asyncon, request, callback, connection = "keep-alive"):
		self.asyncon = asyncon		
		self.wrap_in_chunk = False
		self.end_of_data = False
		self.expect_disconnect = False
		
		self.request = request
		self.callback = callback
		base_request_handler.RequestHandler.__init__ (self, request.logger)
		self.connection = connection				
		
		self.retry_count = 0
		self.http2_handler = None
		
		self.buffer = b""	
		self.response = None
		self.callbacked = 0
		
		self._ssl = isinstance (self.asyncon, asynconnect.AsynSSLConnect)			
		self.method, self.uri = (
			self.request.get_method (),			
			self.asyncon.is_proxy () and self.request.uri or self.request.path
		)
		self.header = []
		if request.get_address () is None:
			request.set_address (self.asyncon.address)
		
	def _del_ (self):
		self.callback = None
		self.asyncon = None
		self.request = None
		self.response = None
		
	#------------------------------------------------
	# handler must provide these methods
	#------------------------------------------------
	def rebuild_http2_headers (self, headers, headers_1x):
		content_encoding = None
		for k, v in headers_1x:
			k = k.lower ()
			if k == "host":
				headers.insert (2, (":authority", v))
				continue
			if k in ("connection", "transfer-encoding"):
				continue
			if k == "content-encodig":
				content_encoding = v
			headers.append ((k, str (v)))
		return headers, content_encoding
	
	def get_http2_upgrade_header (self):
		# try h2c protocol
		hc = {}
		hc ["Upgrade"] = "h2c"
		hc ['Connection'] = 'Upgrade, HTTP2-Settings'
		http2_settings = SettingsFrame (0)
		http2_settings.settings [SettingsFrame.INITIAL_WINDOW_SIZE] = 65535
		settings = base64.urlsafe_b64encode(
		    http2_settings.serialize_body ()
		).rstrip (b'=').decode ("utf8")
		hc ['HTTP2-Settings'] = settings
		return hc
				
	def get_request_header (self, http_version = "1.1", upgrade = True):
		if not self.asyncon.isconnected () and upgrade and http_version == "1.1" and not self._ssl:
			hc = self.get_http2_upgrade_header ()
		elif ((http_version == "1.1" and self.connection == "close") or (http_version == "1.0" and self.connection == "keep-alive")):
			hc = {"Connection": self.connection}
		else:
			hc = {}
			
		auth_header = http_auth.authorizer.make_http_auth_header (self.request, self.asyncon.is_proxy ())
		if http_version == "2.0":
			headers = [
				(":method", self.method),
				(":path", self.uri),
				(":scheme", self._ssl and "https" or "http")
			]
			if auth_header:
				headers.append (("authorization", auth_header))
			return self.rebuild_http2_headers (headers, self.request.get_headers ())
		
		if auth_header:
			hc ["Authorization"] = auth_header
		headers = list (hc.items ()) + self.request.get_headers ()
		self.header = ["%s: %s" % x for x in headers]
		req = ("%s %s HTTP/%s\r\n%s\r\n\r\n" % (
			self.method,
			self.uri,
			http_version,
			"\r\n".join (self.header)
		)).encode ("utf8")		
		return req
		
	def get_request_payload (self):
		return self.request.get_payload ()
						
	def get_request_buffer (self, http_version = "1.1", upgrade = True):
		payload = self.get_request_payload ()
		if type (payload) is bytes:			
			return [self.get_request_header (http_version, upgrade) + payload]
		return [self.get_request_header (http_version, upgrade), payload]	
		
	def collect_incoming_data (self, data):
		if not data:
			self.end_of_data = True
			return
		if not self.response or self.asyncon.get_terminator () == b"\r\n":
			self.buffer += data
		else:
			try:
				self.response.collect_incoming_data (data)			
			except http_response.ContentLimitReached:
				self.asyncon.handle_error (719)
		
	def found_terminator (self):
		if self.response:			
			if self.end_of_data:
				return self.found_end_of_body ()
			
			if self.wrap_in_chunk:
				if self.asyncon.get_terminator () == 0:
					self.asyncon.set_terminator (b"\r\n")
					self.buffer = b""
					return
						
				if not self.buffer:
					return
				
				chunked_size = int (self.buffer.split (b";") [0], 16)				
				self.buffer = b""
				
				if chunked_size == 0:
					self.end_of_data = True
					self.asyncon.set_terminator (b"\r\n")
					
				elif chunked_size > 0:
					self.asyncon.set_terminator (chunked_size)
			
			else:
				self.found_end_of_body ()
						
		else:
			self.expect_disconnect = False
			self.create_response ()
			if not self.response or isinstance (self.response, http_response.FailedResponse):
				return
			
			if self.used_chunk ():
				self.wrap_in_chunk = True
				self.asyncon.set_terminator (b"\r\n") #chunked transfer
			
			else:
				clen = 0 # no transfer-encoding, no content-lenth	
				try:
					clen = self.get_content_length ()
				except TypeError:
					if self.will_be_close ():
						self.expect_disconnect = True
						if self.response.get_header ("content-type"):
							clen = None							
						
				if clen == 0:
					return self.found_end_of_body ()
					
				self.asyncon.set_terminator (clen)
			
	def create_response (self):
		buffer, self.buffer = self.buffer, b""
		try:
			self.response = http_response.Response (self.request, buffer.decode ("utf8"))			
			if self.handle_response_code ():
				return
		except ProtocolSwitchError:
			return self.asyncon.handle_error (716)
		except:
			self.log ("response header error: `%s`" % repr (buffer [:80]), "error")
			return self.asyncon.handle_error (715)			
		
		ct = self.response.check_accept ()
		if ct:	
			self.log ("response content-type error: `%s`" % ct, "error")
			return self.asyncon.handle_close (718)
		
		cl = self.response.check_max_content_length ()
		if cl:
			self.log ("response content-length error: `%d`" % cl, "error")
			return self.asyncon.handle_close (719)
		
	def handle_response_code (self):
		# default header never has "Expect: 100-continue"
		# ignore, wait next message	
		if self.response.code == 100:
			self.asyncon.set_terminator (b"\r\n\r\n")
			self.response = None			
			return 1
			
		elif self.response.code == 101 and self.response.get_header ("Upgrade") == "h2c":	# swiching protocol		
			self.asyncon._proto = "h2c"
			self.response = None
			try:				
				self.switch_to_http2 ()
			except:	
				raise ProtocolSwitchError
			return 1
		return 0	
	
	def handled_http_authorization (self):
		if self.response.code != 401:
			return
						
		if self.request.reauth_count > 0:
			return

		try: 
			http_auth.authorizer.save_http_auth_header (self.request, self.response)
		except:
			self.trace ()
				
	def found_end_of_body (self):
		if self.response:
			self.response.done ()		
		
		self.handled_http_authorization ()			
		if self.will_be_close ():
			self.asyncon.disconnect ()		
		self.close_case_with_end_tran ()
	
	def enter_shutdown_process (self):		
		self.asyncon.handle_close (705)
		
	def connection_closed (self, why, msg):		
		if self.response and self.expect_disconnect:
			self.close_case ()
			return
		
		# possibly disconnected cause of keep-alive timeout
		# but works only HTTP 1.1
		if not self.http2_handler and why == 700 and self.response is None and self.retry_count == 0:
			self.retry_count = 1			
			self.handle_rerequest ()
			return True
		
		self.response = http_response.FailedResponse (why, msg, self.request)
		if hasattr (self.asyncon, "begin_tran"):
			self.close_case ()
			
	def close_case_with_end_tran (self):		
		self.asyncon.end_tran ()
		#print (self.request.meta ['sid'], 'end_tran')
		self.close_case ()
		
	def handle_callback (self):
		if self.callbacked:
			return
		tuple_cb (self, self.callback)
		self.callbacked = 1
	
	def close_case (self):		
		self.handle_callback ()
	
	def switch_to_http2 (self):		
		if self.http2_handler is None:			
			self.http2_handler = http2_request_handler.RequestHandler (self)			
		else:
			self.http2_handler.handle_request (self)	
		
	def has_been_connected (self):
		if self._ssl or self.request.initial_http_version == "2.0":
			if self.request.initial_http_version == "2.0":
				self.asyncon.set_proto ("h2c")			
			if self.asyncon._proto in H2_PROTOCOLS:				
				self.switch_to_http2 ()
			else:
				for data in self.get_request_buffer ("1.1", False):
					self.asyncon.push (data)
	
	def handle_rerequest (self):
		# init for redirecting or reauth
		self.response = None
		self.end_of_data = False
		self.buffer = b''
		self.callbacked = 0
		self.handle_request ()
						
	def handle_request (self):
		if self.asyncon.isconnected () and self.asyncon.get_proto ():			
			#print (self.request.meta ['sid'], 'switch_to_http2')
			return self.switch_to_http2 ()
		
		self.buffer, self.response = b"", None
		self.asyncon.set_terminator (b"\r\n\r\n")
		
		if (self.asyncon.connected) or not (self._ssl or self.request.initial_http_version == "2.0"):
			# IMP: if already connected, it means not http2
			upgrade = True
			if self.FORCE_HTTP_11:
				upgrade = False
			elif self.asyncon.isconnected ():
				upgrade = False
			
			for data in self.get_request_buffer ("1.1", upgrade):
				self.asyncon.push (data)
				
		if self._ssl and self.FORCE_HTTP_11 and self.request.initial_http_version != "2.0":
			self.asyncon.negotiate_http2 (False)
			
		self.asyncon.begin_tran (self)
		#print (self.request.meta ['sid'], 'begin tran')
	
	def will_be_close (self):
		if self.connection == "close": #server misbehavior ex.paxnet
			return True
				
		close_it = True
		connection = self.response.get_header ("connection", "").lower ()
		if self.response.version == "1.1":
			if not connection or connection.find ("keep-alive") != -1 or connection.find ("upgrade") != -1:
				close_it = False
		else:
			if connection.find ("keep-alive") != -1:
				close_it = False
		
		if not close_it:
			keep_alive = self.response.get_header ("keep-alive")
			if keep_alive:
				for each in keep_alive.split (","):
					try: 
						k, v = each.split ("=", 1)
					except ValueError:
						continue
					
					if k.strip () == "timeout": 
						timeout = int (v)
						if timeout < self.asyncon.zombie_timeout:
							self.asyncon.set_timeout (timeout)
					elif k.strip () == "max" and int (v) == 0:
						close_it = True
								
		return close_it
	
	def used_chunk (self):
		transfer_encoding = self.response.get_header ("transfer-encoding")
		return transfer_encoding and transfer_encoding.lower () == "chunked"
	
	def get_content_length (self):
		return int (self.response.get_header ("content-length"))
	
	def get_content_type (self):
		return self.response.get_header ("content-type")
		
	def get_http_version (self):
		return self.response.version
   	
   	
   	


from ..protocols.dns import asyndns
import time
import threading
import os

class DNSCache:
	maintern_interval = 17
	def __init__ (self, prefer_protocol, logger = None):
		self.logger = logger
		self.prefer_protocol = prefer_protocol
		self.__last_maintern = time.time ()
		self.__lock = threading.RLock ()
		self.cache = {}
		self.hits = 0

	def set (self, answers):
		if not answers:
			return
		
		with self.__lock:
			for answer in answers:
				name = answer['name'].lower ()			
				try:
					ttl = int (answer ["ttl"])	
				except (KeyError, ValueError):
					ttl = 300
				answer ["valid"]	= time.time () + max (ttl, 300)
				if name not in self.cache:					
					self.cache [name] = answer
			
	def expire (self, host):
		with self.__lock:
			try: del self.cache [host]
			except KeyError: pass		
	
	def maintern (self, now):
		deletables = []
		with self.__lock:
			for host in self.cache:			
				ans = self.cache [host]
				if ans ["valid"] < now:
					deletables.append (host)					
			for host in deletables:
				del self.cache [host]
		
		self.__last_maintern = time.time ()
			
	def get (self, host, qtype = "A", check_ttl = True):
		now = time.time ()
		if now - self.__last_maintern > self.maintern_interval:
			self.maintern (now)

		host = host.lower ()
		with self.__lock:
			while 1:
				try: 
					answer = self.cache [host]
				except KeyError: 
					return []
				else:
					tn = answer ['typename']
					if tn == "CNAME":
						host = answer ['data'].lower ()
					elif tn == qtype:	
						break
		
		if check_ttl and answer ["valid"] < now:				
			# extend 30 seconds for other querees
			answer ['valid'] = now + (answer ["data"] and 30 or 1)
			# new query will be started
			return []
			
		else:
			return [answer]
	
	def is_ip (self, name):
		arr = name.split (".")
		if len (arr) != 4: return False
		try: arr = [x for x in map (int, arr) if x & 255 == x]
		except ValueError: 
			return False
		if len (arr) != 4: return False
		return True
		
	def __call__ (self, host, qtype, callback):
		self.hits += 1
		hit = self.get (host, qtype, True)		
		if hit: 
			return callback (hit)
		
		if self.is_ip (host):
			self.set ([{"name": host, "data": host, "typename": qtype}])
			return callback ([{"name": host, "data": host, "typename": qtype}])
		
		try:			
			asyndns.query (host, qtype = qtype, protocol = self.prefer_protocol, callback = [self.set, callback])			
		except:
			self.logger.trace (host)
			hit = [{"name": host, "data": None, "typename": qtype, 'ttl': 60}]
			self.set (hit)
			callback (hit)


query = None

def init (logger, dns_servers = [], prefer_protocol = os.name == "nt" and "tcp" or "udp"):
	asyndns.create_pool (dns_servers, logger)
	global query
	if query is None:
		query = DNSCache (prefer_protocol, logger)

def get (name, qtype = "A"):	
	global query
	return query.get (name, qtype, False)



import os
import pickle as translator
from bsddb import db
from . import pathtool

#--------------------------------------------------------------
# DB creation func
#--------------------------------------------------------------
def _checkflag(flag):
	if flag == 'r':
		flags = db.DB_RDONLY
	elif flag == 'rw':
		flags = 0
	elif flag == 'w':
		flags =  db.DB_CREATE
	elif flag == 'c':
		flags =  db.DB_CREATE
	elif flag == 'n':
		flags = db.DB_CREATE | db.DB_TRUNCATE
	else:
		raise error("flags should be one of 'r', 'w', 'c' or 'n'")
	return flags | db.DB_THREAD

def btopen(file, flag='c', mode=0o666,
			btflags=0, cachesize=None, maxkey=None, minkey=None,
			pgsize=None, lorder=None):

	flags = _checkflag(flag)
	d = db.DB()
	if cachesize is not None: d.set_cachesize(0, cachesize)
	if pgsize is not None: d.set_pagesize(pgsize)
	if lorder is not None: d.set_lorder(lorder)
	d.set_flags(btflags)
	if minkey is not None: d.set_bt_minkey (minkey)
	d.open(file, db.DB_BTREE, db.DB_THREAD | flags, mode)   
	return d 


def zbtopen(*arg, **karg):
	return ZObject (btopen (*arg, **karg))
	
	
#--------------------------------------------------------------
# Z Object Handler Wrapper Classes
#--------------------------------------------------------------
class ZCursor:
	def __init__(self, cursor):
		self.dbc = cursor

	def __del__(self):
		self.close()
		
	def __getattr__(self, name):
		return getattr(self.dbc, name)
		
	def dup(self, flags=0):
		return Zcursor(self.dbc.dup(flags))

	def put(self, key, value, flags=0):
		data = translator.dumps(value, 1)
		return self.dbc.put(key, data, flags)

	def get(self, *args):
		count = len(args)  # a method overloading hack
		method = getattr(self, 'get_%d' % count)
		method(*args)

	def get_1(self, flags):
		rec = self.dbc.get(flags)
		return self._extract(rec)

	def get_2(self, key, flags):
		rec = self.dbc.get(key, flags)
		return self._extract(rec)

	def get_3(self, key, value, flags):
		data = translator.dumps(value, 1)
		rec = self.dbc.get(key, flags)
		return self._extract(rec)

	def current(self, flags=0): return self.get_1(flags|db.DB_CURRENT)
	def first(self, flags=0): return self.get_1(flags|db.DB_FIRST)
	def last(self, flags=0): return self.get_1(flags|db.DB_LAST)
	def next(self, flags=0): return self.get_1(flags|db.DB_NEXT)
	def prev(self, flags=0): return self.get_1(flags|db.DB_PREV)
	def consume(self, flags=0): return self.get_1(flags|db.DB_CONSUME)
	def next_dup(self, flags=0): return self.get_1(flags|db.DB_NEXT_DUP)
	def next_nodup(self, flags=0): return self.get_1(flags|db.DB_NEXT_NODUP)
	def prev_nodup(self, flags=0): return self.get_1(flags|db.DB_PREV_NODUP)

	def get_both(self, key, value, flags=0):
		data = translator.dumps(value, 1)
		rec = self.dbc.get_both(key, flags)
		return self._extract(rec)

	def set(self, key, flags=0):
		rec = self.dbc.set(key, flags)
		return self._extract(rec)

	def set_range(self, key, flags=0):
		rec = self.dbc.set_range(key, flags)
		return self._extract(rec)

	def set_recno(self, recno, flags=0):
		rec = self.dbc.set_recno(recno, flags)
		return self._extract(rec)

	set_both = get_both

	def _extract(self, rec):
		if rec is None:
			return None
		else:
			key, data = rec
			return key, translator.loads(data)
			

class ZObject:
	def __init__ (self, db):
		self.db = db
	
	def __len__(self):
		return len(self.db)
		
	def __setitem__ (self, key, value):
		self.db [key] = translator.dumps (value, 1)
		
	def __getattr__ (self, name):	
		return getattr (self.db, name)
	
	def __delitem__ (self, key):
		del self.db [key]
			
	def __getitem__ (self, key):
		return translator.loads (self.db [key])
		
	def values (self):
		return list(map (translator.loads, list(self.db.values ())))
		
	def items (self):
		return [(k, translator.loads (v)) for k, v in list(self.db.items ())]
	
	def get (self, *args, **kw):
		data = self.db.get(*args, **kw)
		if data:
			return translator.loads (data)
		else:
			return None	
		
	def put(self, key, value, txn=None, flags=0):
		data = translator.dumps (value, 1)
		return self.db.put(key, data, txn, flags)
	
	def cursor(self):
		c = ZCursor(self.db.cursor())		
		return c


#--------------------------------------------------------------
# BTREE bsd file group handler
#--------------------------------------------------------------	
class BT:
	def __init__ (self, home):
		self.home = home
		pathtool.mkdir (home)
		self.setup ()
		self.opened = []
		
	def setup (self):
		pass
					
	def open (self, filename, flag="c", zflag = 0, **karg):
		dbfile = os.path.join (self.home, filename)
		
		if zflag:
			d = ZObject (btopen (dbfile, flag, **karg))
		else:
			d = btopen (dbfile, flag, **karg)
		self.opened.append ((d, dbfile))
		return d
		
	def close (self):
		for db, dbfile in self.opened:			
			try: db.close ()				
			except: pass
		
	
	def remove (self):
		for db, dbfile in self.opened:			
			db.remove (dbfile)			

#--------------------------------------------------------------
# ZObject BTREE bsd file group handler
#--------------------------------------------------------------	
class ZBT (BT):
	def open (self, filename, flag="c"):
		return BT.open (self, filename, flag, 1)	
	

#--------------------------------------------------------------
# Environmental BTREE bsd file group handler
#--------------------------------------------------------------	
class EBT (BT):		
	dbopenflags = db.DB_THREAD
	envflags	= db.DB_THREAD | db.DB_INIT_CDB | db.DB_INIT_MPOOL	
	dbtype  = db.DB_BTREE
	dbsetflags   = 0
	
	def setup (self):
		self.env = db.DBEnv()
		self.setEnvOpts()
		self.env.open (self.home, self.envflags | db.DB_CREATE)
	
	def setEnvOpts(self):
		self.env.set_lk_detect(db.DB_LOCK_DEFAULT)
	
	def open (self, filename, flag="c"):
		d = db.DB(self.env)
		if self.dbsetflags:
			d.set_flags(self.dbsetflags)
		d.open(filename, self.dbtype, self.dbopenflags | _checkflag (flag))
		self.opened.append (d)
		return d
	
	def close (self):
		for db in self.opened:
			try: db.close ()
			except: pass
		try: self.env.close ()
		except: pass

		
if __name__ == "__main__":
	d = ZBT ("test")
	z = d.open ("test", "r")
	z ['a'] = [0,1,2,3,4,5,6]
	z ['b'] = {}
	print(z ['a'])
	print(z ['b'])
	print(list(z.keys ()))
	print(list(z.items ()))
	print(list(z.values ()))
	c = z.cursor ()
	while 1:
		r =  next(c)
		if not r: break
		print(r)
		print(c.current ())
	c.close ()	
	d.close ()